#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings_char_3.h"
#include "pairings.h"
#include "funcs.h"


//----------------------------------------------------------------------------
//                      Pairing in characteristic 3
//
// WARNING -- The  *_char_3 functions assume the elliptic curve E is given by
//            an equation of the form :
//            
//                a) y^2 = x^3 + ax^2 + b   if E is not supersingular
//                b) y^2 = x^3 + ax + b     if E is supersingular
//----------------------------------------------------------------------------
static int _is_supersingular_char_3(GEN E) {
    //
    // Returns non zero iif the curve E over a field GF(3^m) is
    // supersingular.
    // WARNING -- Supposes that E is given by y^2 = x^3 + ax + b if
    //            it is supersingular
    //
    if (FF_equal0(gel(E, 1)) && FF_equal0(gel(E, 2)) && FF_equal0(gel(E, 3))) {
        return 1;
    } else {
        return 0;
    }
}
//----------------------------------------------------------------------------
static GEN _eval_vertical_char_3_ss(GEN A, GEN Q) {
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    if (ell_is_inf(A)) {
        return gen_1;
    }
    return gsub(gel(Q, 1), gel(A, 1));
}
//----------------------------------------------------------------------------
static void _eval_paf_tangent_char_3_ss(GEN E, paf* pf, GEN Q) {
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point Q to update pf->num and finally doubles the point
    // pf->point.
    //
    // Given P1 = (x1, y1) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x1
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = a / (2y1)
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    GEN x1 = gel(pf->point, 1);
    GEN y1 = gel(pf->point, 2);
        
    GEN slope = gdiv(gel(E, 4), mului(2, y1));
    
    GEN x3 = gsub(gsub(gsqr(slope), x1), x1);
    GEN y3 = gadd(y1, gmul(slope, gsub(x3, x1)));
    y3 = gneg(y3);
    
    GEN line = gsub(gel(Q, 2), gadd(gmul(slope, gsub(gel(Q, 1), x1)) , y1));
    pf->num = gmul(pf->num, line);

    gel(pf->point, 1) = x3;
    gel(pf->point, 2) = y3;
}
//----------------------------------------------------------------------------
static void _eval_paf_chord_char_3_ss(GEN E, paf* pf, GEN A, GEN Q) {
    //
    // Computes the equation of the line through pf.point and A, evaluates
    // it at the point Q to update pf->num and finally adds A to the point
    // pfunc->point.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P2 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2
    //     y3 = -(y1 + slope * (x3 - x1))
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    // The line (P1-P2) has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    if (_are_same_point(pf->point, A)) {
        return _eval_paf_tangent_char_3_ss(E, pf, Q);
    }
    if (_are_opposite_points(E, pf->point, A)) {
        pf->point = ellinf();
        pf->num  = gmul(pf->num, _eval_vertical_char_3_ss(pf->point, Q));
        return;
    }
    GEN x1 = gel(pf->point, 1);
    GEN y1 = gel(pf->point, 2);
    
    GEN x2 = gel(A, 1);
    GEN y2 = gel(A, 2);
    
    GEN slope  = gdiv(gsub(y2, y1), gsub(x2, x1));
    
    GEN x3 = gsub(gsub(gsqr(slope), x1), x2);
    GEN y3 = gadd(y1, gmul(slope, gsub(x3, x1)));
    y3 = gneg(y3);
    
    GEN line = gsub(gel(Q, 2), gadd(gmul(slope, gsub(gel(Q, 1), x1)) , y1));
    pf->num = gmul(pf->num, line);
    
    gel(pf->point, 1) = x3;
    gel(pf->point, 2) = y3;
}
//----------------------------------------------------------------------------
static GEN _eval_vertical_char_3_nss(GEN A, GEN Q) {
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    if (ell_is_inf(A)) {
        return gen_1;
    }
    return gsub(gel(Q, 1), gel(A, 1));
}
//----------------------------------------------------------------------------
static void _eval_paf_tangent_char_3_nss(GEN E, paf* pf, GEN Q) {
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point Q to update pf->num and finally doubles the point
    // pf->point.
    //
    // Given P1 = (x1, y1) then [2]P2 = (x3, y3) is given by:
    //
    //     x3 = slope^2 - 2.x1 - a
    //     y3 = -(slope * (x3 - x1) + y1)
    //
    // with: slope = a.x1 / y1
    //
    // The line tangent to [2]P has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    GEN x1 = gel(pf->point, 1);
    GEN y1 = gel(pf->point, 2);
        
    GEN slope = gdiv(gmul(gel(E, 2), x1), y1);
    
    GEN x3 = gsub(gsub(gsqr(slope), x1), x1);
    GEN y3 = gadd(y1, gmul(slope, gsub(x3, x1)));
    y3 = gneg(y3);
    
    GEN line = gsub(gel(Q, 2), gadd(gmul(slope, gsub(gel(Q, 1), x1)) , y1));
    pf->num = gmul(pf->num, line);
    
    gel(pf->point, 1) = x3;
    gel(pf->point, 2) = y3;
}
//----------------------------------------------------------------------------
static void _eval_paf_chord_char_3_nss(GEN E, paf* pf, GEN A, GEN Q) {
    //
    // Computes the equation of the line through pf.point and A, evaluates
    // it at the point Q to update pf->num and finally adds A to the point
    // pfunc->point.
    //
    // Given P1 = (x1, y1) and P2 = (x2, y2) then P2 + P2 = (x3, y3)
    // is given by:
    //
    //     x3 = slope^2 - x1 - x2 - a
    //     y3 = -(y1 + slope * (x3 - x1))
    //
    // with: slope = (y2 - y1) / (x2 - x1)
    //
    // The line (P1-P2) has for equation:
    //
    //     Y - slope * (X - x1) - y1
    //
    if (_are_same_point(pf->point, A)) {
        return _eval_paf_tangent_char_3_nss(E, pf, Q);
    }
    if (_are_opposite_points(E, pf->point, A)) {
        pf->point = ellinf();
        pf->num  = gmul(pf->num, _eval_vertical_char_3_nss(pf->point, Q));
        return;
    }
    GEN x1 = gel(pf->point, 1);
    GEN y1 = gel(pf->point, 2);
    
    GEN x2 = gel(A, 1);
    GEN y2 = gel(A, 2);
    
    GEN slope  = gdiv(gsub(y2, y1), gsub(x2, x1));
    
    GEN x3 = gsub(gsub(gsub(gsqr(slope), x1), x2), gel(E, 2));
    GEN y3 = gadd(y1, gmul(slope, gsub(x3, x1)));
    y3 = gneg(y3);
    
    GEN line = gsub(gel(Q, 2), gadd(gmul(slope, gsub(gel(Q, 1), x1)) , y1));
    pf->num = gmul(pf->num, line);
    
    gel(pf->point, 1) = x3;
    gel(pf->point, 2) = y3;
}
//----------------------------------------------------------------------------
static GEN _std_miller_char_3_nss(GEN E, GEN r, GEN P, GEN Q) {

    paf pf;
    pf.num  = gcopy(gen_1);
    pf.point = gcopy(P);

    GEN denom = gcopy(gen_1);
    
    long i = expi(r) - 1;
    while (i >= 0) {
        pf.num = gsqr(pf.num);
        denom   = gsqr(denom);
        
        _eval_paf_tangent_char_3_nss(E, &pf, Q);
        denom = gmul(denom, _eval_vertical_char_3_nss(pf.point, Q));
        
        if (_get_bit(r, i)) {
            _eval_paf_chord_char_3_nss(E, &pf, P, Q);
            denom = gmul(denom, _eval_vertical_char_3_nss(pf.point, Q));
        }
        i--;
    }
    return gdiv(pf.num, denom);
}
//----------------------------------------------------------------------------
static GEN _std_miller_char_3_ss(GEN E, GEN r, GEN P, GEN Q) {

    paf pf;
    pf.num  = gcopy(gen_1);
    pf.point = gcopy(P);

    GEN denom = gcopy(gen_1);
    
    long i = expi(r) - 1;
    while (i >= 0) {
        pf.num = gsqr(pf.num);
        denom   = gsqr(denom);
        
        _eval_paf_tangent_char_3_ss(E, &pf, Q);
        denom = gmul(denom, _eval_vertical_char_3_ss(pf.point, Q));
        
        if (_get_bit(r, i)) {
            _eval_paf_chord_char_3_ss(E, &pf, P, Q);
            denom = gmul(denom, _eval_vertical_char_3_ss(pf.point, Q));
        }
        i--;
    }
    return gdiv(pf.num, denom);
}
//----------------------------------------------------------------------------
static GEN _std_miller_char_3(GEN E, GEN r, GEN P, GEN Q) {
    if (_is_supersingular_char_3(E)) {
        return _std_miller_char_3_ss(E, r, P, Q);
    }
    return _std_miller_char_3_nss(E, r, P, Q);
}
//----------------------------------------------------------------------------
GEN _tate_pairing_char_3(GEN E, GEN P, GEN Q, GEN r)
{
    pari_sp ltop = avma;
    
    checkellpt(P);
    checkellpt(Q);
    
    if (!ellisoncurve(E, P) || !ellisoncurve(E, Q)) {
        ERR("point(s) not on curve");
    }
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }
    GEN tate = _std_miller_char_3(E, r, P, Q);
    
    GEN  p = FF_p(gel(P, 1));
    long k = degpol(FF_mod(gel(Q, 1)));
    
    tate = FF_pow(tate, diviiexact(subis(powiu(p, k), 1), r));

    return gerepilecopy(ltop, tate);
}
//----------------------------------------------------------------------------
void update_pairing_data_char_3(pairing_data_t* pdata) {
    pairing_data_t* dummy = pdata;
    dummy = NULL;
    return;
}
//----------------------------------------------------------------------------