#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings.h"
#include "funcs.h"
#include "commons.h"
#include "jacobian.h"
#include "stats.h"
#include "arithm.h"

//----------------------------------------------------------------------------
//                      Pairing in characteristic > 3
//
// WARNING: All functions in this file assume the elliptic curve is given
//          by an equation of the form y^2 = x^3 + ax + b
//
// WARNING: Excepted when mentionned otherwise, all the functions in this
//          file are NOT memory safe!
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_jac_fp(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_p).
    //
    // Coordinates of Q and resulting point are given by t_INTs.
    // p is a t_INT.
    //
    return mkvecn(3, gel(Q, 1), FP_NEG(gel(Q, 2), p));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
_get_opposite_point_jac_fpxq(GEN Q, GEN p) {
    //
    // Returns the opposite of the point Q on E(F_{p^k}).
    //
    // Coordinates of Q and resulting point are given by ZXs.
    // p is a t_INT.
    //
    return mkvecn(3, gel(Q, 1), FPX_NEG(gel(Q, 2), p), gel(Q, 1));
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_jac_fp(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN p) {
    IN();
    //
    // Return true if the points A and B on E(F_p) are opposites.
    // The coordinates of A and B are given by t_INTs.
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = gel(A, 1);
    GEN y1 = gel(A, 2);
    GEN z1 = gel(A, 3);
    
    GEN x2 = gel(B, 1);
    GEN y2 = gel(B, 2);
    GEN z2 = gel(B, 3);
    
    GEN z12 = SQRI(z1);
    GEN z22 = SQRI(z2);
    
    GEN x1z22 = FP_MUL(x1, z22, p);
    GEN x2z12 = FP_MUL(x2, z12, p);
    
    GEN z13 = MULII(z12, z1);
    GEN z23 = MULII(z22, z2);
    
    GEN y1z23 = FP_MUL(y1, z23, p);
    GEN y2z13 = FP_MUL(y2, z13, p);
    
    if (!equalii(x1z22, x2z12)) {
        OUT();
        return 0;
    }
    if (equalii(y1z23, FP_NEG(y2z13, p))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE int
_are_opposite_points_jac_fpxq(MAYBE_UNUSED GEN E, GEN A, GEN B, GEN t, GEN p) {
    IN();
    //
    // Return true if the points A and B on E(F_{p^k}) are opposites.
    // The coordinates of A and B are given by FpXQs (i.e. Z_Xs).
    // t is a FpXQ (i.e. Z_X).
    // p is a t_INT.
    //
    // WARNING -- Supposes the curve is given by y^2 = x^3 + a.x + b
    //
    int a_is_inf = ell_is_inf(A);
    int b_is_inf = ell_is_inf(B);

    if (a_is_inf && b_is_inf) {
        OUT();
        return 1;
    }
    if (a_is_inf || b_is_inf) {
        OUT();
        return 0;
    }
    GEN x1 = gel(A, 1);
    GEN y1 = gel(A, 2);
    GEN z1 = gel(A, 3);
    
    GEN x2 = gel(B, 1);
    GEN y2 = gel(B, 2);
    GEN z2 = gel(B, 3);
    
    GEN z12 = FPXQ_SQR(z1, t, p);
    GEN z22 = FPXQ_SQR(z2, t, p);
    
    GEN x1z22 = FPXQ_MUL(x1, z22, t, p);
    GEN x2z12 = FPXQ_MUL(x2, z12, t, p);
    
    GEN z13 = FPXQ_MUL(z12, z1, t, p);
    GEN z23 = FPXQ_MUL(z22, z2, t, p);
    
    GEN y1z23 = FPXQ_MUL(y1, z23, t, p);
    GEN y2z13 = FPXQ_MUL(y2, z13, t, p);
    
    if (!ZX_equal(x1z22, x2z12)) {
        OUT();
        return 0;
    }
    if (ZX_equal(y1z23, FPX_NEG(y2z13, p))) {
        OUT();
        return 1;
    }
    OUT();
    return 0;
}
//----------------------------------------------------------------------------
GEN mul_jac_fp(pairing_data_t* pdata, GEN a, GEN P) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AP = gcopy(P);
    GEN MP = _get_opposite_point_jac_fp(P, pdata->charac);
    
    while (i >= 0) {
        AP = double_jac_fp(pdata, AP);
        switch (naf[i]) {
            case 1:
                AP = add_jac_fp(pdata, AP, P);
                break;
            case -1:
                AP = add_jac_fp(pdata, AP, MP);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);
    return AP;
}
//----------------------------------------------------------------------------
GEN add_jac_fp(pairing_data_t* pdata, GEN P1, GEN P2) {
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, z2) then P1 + P2 = (x3, y3, z3)
    // is given by:
    // 
    //    x3 = r^2 - j - 2*v
    //    y3 = r*(v-x3) - 2*s1*j
    //    z3 = ((z1+z2)^2 - z1z1 - z2z2)*h
    //
    // with:
    //
    //    z1z1 = z1^2
    //    z2z2 = z2^2
    //    u1 = x1*z2z2
    //    u2 = x2*z1z1
    //    s1 = y1*z2*z2z2
    //    s2 = y2*z1*z1z1
    //    h  = u2-u1
    //    i  = (2*h)2
    //    j  = h*i
    //    r  = 2*(s2-s1)
    //    v  = u1*i
    // 
    // (This is the "add-2007-bl" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    if (ell_is_inf(P1)) {
        return gcopy(P2);
    }
    if (ell_is_inf(P2)) {
        return gcopy(P1);
    }
    GEN p = pdata->charac;
    
    GEN x1 = gel(P1, 1);
    GEN y1 = gel(P1, 2);
    GEN z1 = gel(P1, 3);
    
    GEN x2 = gel(P2, 1);
    GEN y2 = gel(P2, 2);
    GEN z2 = gel(P2, 2);

    GEN z1z1   = FP_SQR(z1, p);
    GEN z1z1z1 = MULII(z1, z1z1);
    GEN z2z2   = FP_SQR(z2, p);
    GEN z2z2z2 = MULII(z2, z2z2);
    
    GEN u1 = REM(MULII(x1, z2z2), p);
    GEN u2 = REM(MULII(x2, z1z1), p);
    GEN s1 = REM(MULII(y1, z2z2z2), p);
    GEN s2 = REM(MULII(y2, z1z1z1), p);
    
    if (equalii(u2, u2)) {      
        if (equalii(s2, s1)) {
            return double_jac_fp(pdata, P1);
        } else {
            return ellinf();
        }
    }
    GEN h  = SUBII(u2, u1);
    GEN hh = FP_SQR(h, p);
    GEN i  = MULUI(4, hh);
    GEN j  = MULII(h, i);
    GEN r  = MULUI(2, SUBII(s2, s1));
    GEN v  = MULII(u1, i);
    
    GEN x3 = SUBII(SUBII(SQRI(r), j), MULUI(2, v));
    GEN y3 = SUBII(MULII(r, SUBII(v, x3)), MULII(MULUI(2, s1), j));
    GEN z3 = MULII(h, SUBII(SUBII(SQRI(ADDII(z1, z2)), z1z1), z2z2));
    
    x3 = Rg_to_Fp(x3, p);
    y3 = Rg_to_Fp(y3, p);
    z3 = Rg_to_Fp(z3, p);

    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN double_jac_fp(pairing_data_t* pdata, GEN P) {
    //
    // Given P = (x1, y1, z1) then [2]P = (x3, y3, z3) is given by:
    //
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    //
    // with:
    //
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    //
    // (This is the "dbl-2007-bl" doubling formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    if (ell_is_inf(P)) {
        return ellinf();
    } 
    GEN x1 = gel(P, 1);
    GEN y1 = gel(P, 2);
    GEN z1 = gel(P, 3);
    
    if (cmpis(y1, 0) == 0) {
        return ellinf();
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN a = gel(gel(E, 4), 2);
    
    GEN xx   = FP_SQR(x1, p);
    GEN yy   = FP_SQR(y1, p);
    GEN zz   = FP_SQR(z1, p);    
    GEN yyyy = FP_SQR(yy, p);
    
    GEN s = MULUI(2, SUBII(SUBII(SQRI(ADDII(x1, yy)), xx), yyyy));
    GEN m = ADDII(MULUI(3, xx), MULII(a, SQRI(zz)));
    GEN u = SUBII(SQRI(m), MULUI(2, s));
    
    GEN x3 = u;
    GEN y3 = SUBII(MULII(m, SUBII(s, u)), MULUI(8, yyyy));
    GEN z3 = SUBII(SUBII(SQRI(ADDII(y1, z1)), yy), zz);

    x3 = Rg_to_Fp(x3, p);
    y3 = Rg_to_Fp(y3, p);
    z3 = Rg_to_Fp(z3, p);

    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN mul_jac_fpxq(pairing_data_t* pdata, GEN a, GEN Q) {
    int* naf = _naf(a);
    int  i = expi(a) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;
    
    GEN AQ = gcopy(Q);
    GEN MQ = _get_opposite_point_jac_fpxq(Q, pdata->charac);
    
    while (i >= 0) {
        AQ = double_jac_fpxq(pdata, AQ);
        switch (naf[i]) {
            case 1:
                AQ = add_jac_fpxq(pdata, AQ, Q);
                break;
            case -1:
                AQ = add_jac_fpxq(pdata, AQ, MQ);
                break;
            default:
                break;
        }
        i--;
    }
    free(naf);
    return AQ;
}
//----------------------------------------------------------------------------
GEN add_jac_fpxq(pairing_data_t* pdata, GEN Q1, GEN Q2) {
    // 
    //    x3 = r^2 - j - 2*v
    //    y3 = r*(v-x3) - 2*s1*j
    //    z3 = ((z1+z2)^2 - z1z1 - z2z2)*h
    //
    // with:
    //
    //    z1z1 = z1^2
    //    z2z2 = z2^2
    //    u1 = x1*z2z2
    //    u2 = x2*z1z1
    //    s1 = y1*z2*z2z2
    //    s2 = y2*z1*z1z1
    //    h  = u2-u1
    //    i  = (2*h)2
    //    j  = h*i
    //    r  = 2*(s2-s1)
    //    v  = u1*i
    // 
    // (This is the "add-2007-bl" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    if (ell_is_inf(Q1)) {
        return gcopy(Q2);
    }
    if (ell_is_inf(Q2)) {
        return gcopy(Q1);
    }
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(Q1, 1);
    GEN y1 = gel(Q1, 2);
    GEN z1 = gel(Q1, 3);
    
    GEN x2 = gel(Q2, 1);
    GEN y2 = gel(Q2, 2);
    GEN z2 = gel(Q2, 3);
    
    GEN z1z1   = FPXQ_SQR(z1, t, p);
    GEN z1z1z1 = ZX_MUL(z1, z1z1);
    GEN z2z2   = FPXQ_SQR(z2, t, p);
    GEN z2z2z2 = ZX_MUL(z2, z2z2);
    
    GEN u1 = FPXQ_MUL(x1, z2z2, t, p);
    GEN u2 = FPXQ_MUL(x2, z1z1, t, p);
    GEN s1 = FPXQ_MUL(y1, z2z2z2, t, p);
    GEN s2 = FPXQ_MUL(y2, z1z1z1, t, p);
    
    if (_ZX_equal(u2, u1)) {
        if (_ZX_equal(s2, s1)) {
            return double_jac_fpxq(pdata, Q1);
        } else {
            return ellinf();
        }
    }
    GEN h  = ZX_SUB(u2, u1);
    GEN hh = FPXQ_SQR(h, t, p);
    GEN i  = ZX_MULS(hh, 4);
    GEN j  = FPX_MUL(h, i, p);
    GEN r  = ZX_MULS(ZX_SUB(s2, s1), 2);
    GEN v  = FPX_MUL(u1, i, p);
    
    GEN x3 = ZX_SUB(ZX_SUB(FPXQ_SQR(r, t, p), j), ZX_MULS(v, 2));
    GEN y3 = ZX_SUB(
                FPXQ_MUL(r, ZX_SUB(v, x3), t, p),
                FPXQ_MUL(ZX_MULS(s1, 2), j, t, p)
             );
    GEN z3 = ZX_MUL(h, ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(z1, z2), t, p), z1z1), z2z2));
    
    x3 = Rg_to_FpXQ(x3, t, p);
    y3 = Rg_to_FpXQ(y3, t, p);
    z3 = Rg_to_FpXQ(z3, t, p);
 
    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN double_jac_fpxq(pairing_data_t* pdata, GEN Q) {
    //
    // Given Q = (x1, y1, z1) then [2]Q = (x3, y3, z3) is given by:
    //
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    //
    // with:
    //
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    //
    IN();
    if (ell_is_inf(Q)) {
        return ellinf();
    }     
    GEN x1 = gel(Q, 1);
    GEN y1 = gel(Q, 2);
    GEN z1 = gel(Q, 3);
    
    if (signe(y1) == 0) {
        return ellinf();
    }
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN a = gel(gel(E, 4), 2);
    
    GEN xx   = FPXQ_SQR(x1, t, p);
    GEN yy   = FPXQ_SQR(y1, t, p);
    GEN zz   = FPXQ_SQR(z1, t, p);
    GEN yyyy = FPXQ_SQR(yy, t, p);
    
    GEN s = ZX_MULS(ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(x1, yy), t, p), xx), yyyy),2);
    GEN m = ZX_ADD(ZX_MULS(xx, 3), FPX_FP_MUL(FPXQ_SQR(zz, t, p), a, p));
    GEN u = ZX_SUB(FPXQ_SQR(m, t, p), ZX_MULS(s, 2));
    
    GEN x3 = u;
    GEN y3 = ZX_SUB(ZX_MUL(m, ZX_SUB(s, u)), ZX_MULS(yyyy, 8));
    GEN z3 = ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(y1, z1), t, p), yy), zz);

    x3 = Rg_to_FpXQ(x3, t, p);
    y3 = Rg_to_FpXQ(y3, t, p);
    z3 = Rg_to_FpXQ(z3, t, p);

    return mkvecn(3, x3, y3, z3);
}
//----------------------------------------------------------------------------
GEN
eval_vert_jac_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    // Supposes that Q = (X, Y, 1).
    //
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, t, p);
    }
    GEN V = FPX_FP_MUL(gel(Q, 1), FP_SQR(gel(A, 3), p), p);
    V = FPX_FP_ADD(V, FP_NEG(gel(A, 1), p), p);
    
    OUT();
    return V;
}
//----------------------------------------------------------------------------
void
update_vert_jac_fp_fpxq(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // A is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // Return value is a ZX.
    //
    // Supposes that Q = (X, Y, 1).
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        GEN t = pdata->irredpol;

        GEN V = FPX_FP_MUL(gel(Q, 1), FP_SQR(gel(A, 3), p), p);
        V = FPX_FP_ADD(V, FP_NEG(gel(A, 1), p), p);

        pdata->denom = FPXQ_MUL(pdata->denom, V, t, p);
        OUT();
        return;
    }
}
//----------------------------------------------------------------------------
void
update_tangent_jac_fp_fpxq(pairing_data_t* pdata, GEN Q)
{
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point Q to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_p) whose coordinates are given by t_INTs.
    // Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P = (x1, y1, z1) then [2]P = (x3, y3, z3) is given by:
    //
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    //
    // with:
    //
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    //
    // (This is the "dbl-2007-bl" doubling formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    // The line tangent to [2]P has for equation:
    //
    //     L(PP) : Y - slope * (X - x1/z1^2) - y1/z1^3  (= 0)
    //
    // with: slope = m / (2 * y * z)
    //
    // The computed line here is taken as:
    //
    //    CL(PP) = 2.y1.z1^3.L(PP)
    //           = 2.y1.z1^3.Y - m.(z1^2.X - x1) - 2.y1^2
    //
    // We suppose that the line is evaluated at (XQ, YQ, 1). For Tate-like 
    // pairings the 2.y1.z1^3 will cancel out during the final exponentation.
    //
    IN();
    if (ell_is_inf(pdata->point)) {    
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);
    GEN z1 = gel(pdata->point, 3);
    
    if (cmpis(y1, 0) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        pdata->num = FPXQ_MUL(
                       pdata->num,
                       eval_vert_jac_fp_fpxq(pdata, pdata->point, Q),
                       t, p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN a = gel(gel(E, 4), 2);
    
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    GEN xx   = FP_SQR(x1, p);
    GEN yy   = FP_SQR(y1, p);
    GEN zz   = FP_SQR(z1, p);
    
    GEN yyyy = FP_SQR(yy, p);
    GEN zzz  = MULII(z1, zz);
    
    GEN s = MULUI(2, SUBII(SUBII(SQRI(ADDII(x1, yy)), xx), yyyy));
    GEN m = ADDII(MULUI(3, xx), MULII(a, SQRI(zz)));
    GEN u = SUBII(SQRI(m), MULUI(2, s));
    
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    GEN x3 = u;
    GEN y3 = SUBII(MULII(m, SUBII(s, u)), MULUI(8, yyyy));
    GEN z3 = SUBII(SUBII(SQRI(ADDII(y1, z1)), yy), zz);

    //
    //    CL(PP) = 2.y1.z1^3.L(PP)
    //           = 2.y1.zzz.Y - m.(zz.X - x1) - 2.yy
    //           = 2.y1.zzz.Y - m.zz.X + m.x1 - 2.yy
    //           = la - lb + lc -ld
    //
    GEN la = FPX_FP_MUL(gel(Q, 2), MULUI(2, MULII(y1, zzz)), p);    
    GEN lb = FPX_FP_MUL(gel(Q, 1), MULII(m, zz), p);
    GEN lc = MULII(x1, m);
    GEN ld = MULUI(2, yy);

    GEN line = FPX_FP_ADD(
                 FPX_ADD(la, FPX_NEG(lb, p), p),
                 SUBII(lc, ld),
                 p
               );
    
    gel(pdata->point, 1) = Rg_to_Fp(x3, p);
    gel(pdata->point, 2) = Rg_to_Fp(y3, p);
    gel(pdata->point, 3) = Rg_to_Fp(z3, p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
    
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                         pdata->denom,
                         eval_vert_jac_fp_fpxq(pdata, pdata->point, Q),
                         t,
                         p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_jac_fp_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Computes the equation of the line through pf.point and P (supposed to
    // be different from the point at infinity), evaluates it at the point Q
    // to update pf->num, adds P to the point pfunc->point and finally
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point Q.
    //
    // pf.point and P are points on E(F_p) whose coordinates are given by 
    // t_INTs. Q is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, 1) then P1 + P2 = (x3, y3, z3)
    // is given by:
    // 
    //    x3 = r^2 - j - 2 * v
    //    y3 = r * (v - x3) - 2 * y1 * j
    //    z3 = (z1 + h)^2 - z1z1 - hh
    //
    // with:
    //
    //    z1z1 = z1^2
    //    u2   = x2 * z1z1
    //    s2   = y2 * z1 * z1z1
    //    h    = u2 - x1
    //    hh   = h^2
    //    i    = 4 * hh
    //    j    = h * i
    //    r    = 2 * (s2 - y1)
    //    v    = x1 * i
    //
    // (This is the "madd-2007-bl" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    // The line L(P1P2) through P1 and P2 has for equation:
    //
    //     L(P1P2) : Y - slope * (X - x2) - y2  (= 0) (remember z2 = 1)
    //
    // with: slope = (s2 - y1) / z1.h
    //
    // The computed line here is taken as:
    //
    //    CL(P1P2) = z1.h.L(P1P2)
    //             = z1.h.(Y - y2) - (s2 - y1).(X - x2)
    //
    // We suppose that the line is evaluated at (XQ, YQ, 1). For Tate-like 
    // pairings the v will cancel out during the final exponentation.
    //
    IN();
    
    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
        
    int t_is_inf = ell_is_inf(T);
    if (t_is_inf) {
        //
        // The chord is the vertical line passing through T
        //
        pdata->num = FPXQ_MUL(
                         pdata->num,
                         eval_vert_jac_fp_fpxq(pdata, T, Q),
                         t,
                         p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    //
    // From now on we know that both T and P are not the point at inifinity.
    //
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);
    GEN z1 = gel(T, 3);
    
    GEN x2 = gel(P, 1);
    GEN y2 = gel(P, 2);

    GEN z1z1   = FP_SQR(z1, p);
    GEN z1z1z1 = MULII(z1, z1z1);
    
    GEN u2 = REM(MULII(x2, z1z1), p);
    GEN s2 = REM(MULII(y2, z1z1z1), p);
    
    if (equalii(u2, x1)) {
        //
        // T and P are on the same vertical line.
        //      
        if (equalii(s2, y1)) {
            //
            // T and P are the same point and are different from the point
            // at inifinity.
            //
            update_tangent_jac_fp_fpxq(pdata, Q);
            OUT();
            return;
            
        } else {
            //
            // The chord is the vertical line passing through T
            //
            pdata->num = FPXQ_MUL(
                             pdata->num,
                             eval_vert_jac_fp_fpxq(pdata, T, Q),
                             t,
                             p
                         );
            pdata->point = ellinf();
            OUT();
            return;
        }
    }
    //
    //    x3 = r^2 - j - 2 * v
    //    y3 = r * (v - x3) - 2 * y1 * j
    //    z3 = (z1 + h)^2 - z1z1 - hh
    //
    // with:
    //
    //    h    = u2 - x1
    //    hh   = h^2
    //    i    = 4 * hh
    //    j    = h * i
    //    r    = 2 * (s2 - y1)
    //    v    = x1 * i
    //
    GEN h  = SUBII(u2, x1);
    GEN hh = FP_SQR(h, p);
    GEN i  = MULUI(4, hh);
    GEN j  = MULII(h, i);
    GEN r  = MULUI(2, SUBII(s2, y1));
    GEN v  = MULII(x1, i);
    
    GEN x3 = SUBII(SUBII(SQRI(r), j), MULUI(2, v));
    GEN y3 = SUBII(MULII(r, SUBII(v, x3)), MULII(MULUI(2, y1), j));
    GEN z3 = SUBII(SUBII(SQRI(ADDII(z1, h)), z1z1), hh);
    
    //
    //    CL(P1P2) = z1.h.L(P1P2)
    //              = z1.h.(Y - y2) - (s2 - y1).(X - x2)
    //              = la - lb
    //
    GEN z1h = MULII(z1, h);

    GEN la = FPX_FP_ADD(gel(Q, 2), FP_NEG(y2, p), p);
        la = FPX_FP_MUL(la, z1h, p);
    
    GEN lb = FPX_FP_ADD(gel(Q, 1), FP_NEG(x2, p), p);
        lb = FPX_FP_MUL(lb, SUBII(s2, y1), p);
    
    GEN line = FPX_ADD(la, FPX_NEG(lb, p), p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = Rg_to_Fp(x3, p);
    gel(pdata->point, 2) = Rg_to_Fp(y3, p);
    gel(pdata->point, 3) = Rg_to_Fp(z3, p);
        
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                           pdata->denom,
                           eval_vert_jac_fp_fpxq(pdata, pdata->point, Q),
                           t,
                           p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
GEN
eval_vert_jac_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Supposes that Q = (X, Y, 1).
    //
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    if (ell_is_inf(A)) {
        OUT();
        return Rg_to_FpXQ(gen_1, t, p);
    }
    
    GEN V = FPX_FP_MUL(FPXQ_SQR(gel(A, 3), t, p), gel(Q, 1), p);
    V = FPX_ADD(V, FPX_NEG(gel(A, 1), p), p);
    
    OUT();
    return V;
}
//----------------------------------------------------------------------------
void
update_vert_jac_fpxq_fp(pairing_data_t* pdata, GEN A, GEN Q)
{
    IN();
    //
    // Computes the equation of the vertical line through A and returns
    // its evaluation at the point Q.
    //
    // pf.point and A are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. Q is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Supposes that Q = (X, Y, 1).
    //
    if (pdata->do_denom_elim) {
        return;
    }
    if (!ell_is_inf(A)) {
        GEN p = pdata->charac;
        GEN t = pdata->irredpol;
        
        GEN V = FPX_FP_MUL(FPXQ_SQR(gel(A, 3), t, p), gel(Q, 1), p);
        V = FPX_ADD(V, FPX_NEG(gel(A, 1), p), p);
        
        pdata->denom = FPXQ_MUL(pdata->denom, V, t, p);
        OUT();
        return;
    }
    OUT();
}
//----------------------------------------------------------------------------
void
update_tangent_jac_fpxq_fp(pairing_data_t* pdata, GEN P)
{
    //
    // Computes the equation of the line tangent to pf.point, evaluates
    // it at the point P to update pf->num and finally doubles the point
    // pf->point.
    //
    // pf.point is a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P = (x1, y1, z1) then [2]P = (x3, y3, z3) is given by:
    //
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    //
    // with:
    //
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    //
    // (This is the "dbl-2007-bl" doubling formula from the explicit formula 
    //  database at http://www.hyperelliptic.org/EFD)
    //
    // The line tangent to [2]P has for equation:
    //
    //     L(PP) : Y - slope * (X - x1/z1^2) - y1/z1^3  (= 0)
    //
    // with: slope = m / (2 * y * z)
    //
    // The computed line here is taken as:
    //
    //    CL(PP) = 2.y1.z1^3.L(PP)
    //           = 2.y1.z1^3.Y - m.(z1^2.X - x1) - 2.y1^2
    //
    // We suppose that the line is evaluated at (XP, YP, 1). For Tate-like 
    // pairings the 2.y1.z1^3 will cancel out during the final exponentation.
    //
    IN();
    if (ell_is_inf(pdata->point)) {    
        //
        // The evaluation of the vertical line (OO) is 1 (where O is the point
        // at infinity).
        //
        OUT();
        return;
    } 
    
    GEN E = pdata->curve;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN x1 = gel(pdata->point, 1);
    GEN y1 = gel(pdata->point, 2);
    GEN z1 = gel(pdata->point, 3);
    
    if (signe(y1) == 0) {
        //
        // The tangent is the vertical line passing through x1
        //
        pdata->num = FPXQ_MUL(
                       pdata->num,
                       eval_vert_jac_fp_fpxq(pdata, pdata->point, P),
                       t, p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    GEN a = gel(gel(E, 4), 2);
    
    //     xx   = x1^2
    //     yy   = y1^2
    //     yyyy = yy^2
    //     zz   = z1^2
    //     s    = 2 * ((x1 + yy)^2 - xx - yyyy)
    //     m    = 3 * xx + a * zz^2
    //     u    = m^2 - 2 * s
    GEN xx   = FPXQ_SQR(x1, t, p);
    GEN yy   = FPXQ_SQR(y1, t, p);
    GEN zz   = FPXQ_SQR(z1, t, p);
    GEN yyyy = FPXQ_SQR(yy, t, p);
    GEN zzz  = FPXQ_MUL(z1, zz, t, p);
    
    GEN s = ZX_MULS(
                ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(x1, yy), t, p), xx), yyyy),
                2
            );
    GEN m = ZX_ADD(ZX_MULS(xx, 3), FPX_FP_MUL(FPXQ_SQR(zz, t, p), a, p));
    GEN u = ZX_SUB(FPXQ_SQR(m, t, p), ZX_MULS(s, 2));
    
    //     x3 = u
    //     y3 = m * (s - u) - 8 * yyyy
    //     z3 = (y1 + z1)^2 - yy - zz
    GEN x3 = u;
    GEN y3 = ZX_SUB(ZX_MUL(m, ZX_SUB(s, u)), ZX_MULS(yyyy, 8));
    GEN z3 = ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(y1, z1), t, p), yy), zz);

    //
    //    CL(PP) = 2.y1.z1^3.L(PP)
    //           = 2.y1.zzz.Y - m.(zz.X - x1) - 2.yy
    //           = 2.y1.zzz.Y - m.zz.X + m.x1 - 2.yy
    //           = la - lb + lc -ld
    //
    GEN la = FPX_FP_MUL(ZX_MULS(FPXQ_MUL(y1, zzz, t, p), 2), gel(P, 2) , p);
    GEN lb = FPX_FP_MUL(FPX_MUL(m, zz, p), gel(P, 1), p);
    GEN lc = FPX_MUL(x1, m, p);
    GEN ld = ZX_MULS(yy, 2);

    GEN line = FPX_ADD(
                 FPX_ADD(la, FPX_NEG(lb, p), p),
                 FPX_ADD(lc, FPX_NEG(ld, p), p),
                 p
               );
    
    gel(pdata->point, 1) = Rg_to_FpXQ(x3, t, p);
    gel(pdata->point, 2) = Rg_to_FpXQ(y3, t, p);
    gel(pdata->point, 3) = Rg_to_FpXQ(z3, t, p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);
    
    if (!pdata->do_denom_elim) {
        GEN v = eval_vert_jac_fpxq_fp(pdata, pdata->point, P);
        pdata->denom = FPXQ_MUL(pdata->denom, v, t, p);
    }    
    OUT();
}
//----------------------------------------------------------------------------
void
update_chord_jac_fpxq_fp(pairing_data_t* pdata, GEN Q, GEN P)
{
    //
    // Computes the equation of the line through pf.point and Q (supposed to
    // be different from the point at infinity), evaluates it at the point P
    // to update pf->num, adds Q to the point pfunc->point and finally
    // multiplies pf->denom by the equation of the vertical line at
    // pf->point evaluated at the point P.
    //
    // pf.point and Q are points on E(F_{p^k}) whose coordinates are given by 
    // ZXs. P is a point on E(F_p) whose coordinates are given by t_INTs.
    //
    // Given P1 = (x1, y1, z1) and P2 = (x2, y2, 1) then P1 + P2 = (x3, y3, z3)
    // is given by:
    //
    //    x3 = r^2 - j - 2 * v
    //    y3 = r * (v - x3) - 2 * y1 * j
    //    z3 = (z1 + h)^2 - z1z1 - hh
    //
    // with:
    //
    //    z1z1 = z1^2
    //    u2   = x2 * z1z1
    //    s2   = y2 * z1 * z1z1
    //    h    = u2 - x1
    //    hh   = h^2
    //    i    = 4 * hh
    //    j    = h * i
    //    r    = 2 * (s2 - y1)
    //    v    = x1 * i
    //
    // (This is the "madd-2007-bl" addition formula, as given by the explicit 
    //  formula database at http://www.hyperelliptic.org/EFD)
    //
    // The line L(P1P2) through P1 and P2 has for equation:
    //
    //     L(P1P2) : Y - slope * (X - x2) - y2  (= 0) (remember z2 = 1)
    //
    // with: slope = (s2 - y1) / z1.h
    //
    // The computed line here is taken as:
    //
    //    CL(P1P2) = z1.h.L(P1P2)
    //             = z1.h.(Y - y2) - (s2 - y1).(X - x2)
    //
    // We suppose that the line is evaluated at (XQ, YQ, 1). For Tate-like 
    // pairings the v will cancel out during the final exponentation.
    //
    IN();

    GEN T = pdata->point;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;

    int t_is_inf = ell_is_inf(T);
    if (t_is_inf) {
        //
        // The chord is the vertical line passing through T
        //
        pdata->num = FPXQ_MUL(
                         pdata->num,
                         eval_vert_jac_fpxq_fp(pdata, T, P),
                         t,
                         p
                     );
        pdata->point = ellinf();
        OUT();
        return;
    }
    //
    // From now on we know that both T and Q are not the point at inifinity.
    //
    GEN x1 = gel(T, 1);
    GEN y1 = gel(T, 2);
    GEN z1 = gel(T, 3);
    
    GEN x2 = gel(Q, 1);
    GEN y2 = gel(Q, 2);

    GEN z1z1   = FPXQ_SQR(z1, t, p);
    GEN z1z1z1 = ZX_MUL(z1, z1z1);
    
    GEN u2 = FPXQ_MUL(x2, z1z1, t, p);
    GEN s2 = FPXQ_MUL(y2, z1z1z1, t, p);
    
    if (_ZX_equal(u2, x1)) {
        //
        // T and P are on the same vertical line.
        //
        if (_ZX_equal(s2, y1)) {
            //
            // T and Q are the same point and are different from the point
            // at inifinity.
            //
            update_tangent_jac_fpxq_fp(pdata, P);
            OUT();
            return;
            
        } else {
            //
            // The chord is the vertical line passing through T
            //
            pdata->num = FPXQ_MUL(
                             pdata->num,
                             eval_vert_jac_fpxq_fp(pdata, T, P),
                             t,
                             p
                         );
            pdata->point = ellinf();
            OUT();
            return;
        }
    }
    //
    //    x3 = r^2 - j - 2 * v
    //    y3 = r * (v - x3) - 2 * y1 * j
    //    z3 = (z1 + h)^2 - z1z1 - hh
    //
    // with:
    //
    //    h    = u2 - x1
    //    hh   = h^2
    //    i    = 4 * hh
    //    j    = h * i
    //    r    = 2 * (s2 - y1)
    //    v    = x1 * i
    //
    GEN h  = ZX_SUB(u2, x1);
    GEN hh = FPXQ_SQR(h, t, p);
    GEN i  = ZX_MULS(hh, 4);
    GEN j  = FPX_MUL(h, i, p);
    GEN r  = ZX_MULS(ZX_SUB(s2, y1), 2);
    GEN v  = FPX_MUL(x1, i, p);
    
    GEN x3 = ZX_SUB(ZX_SUB(FPXQ_SQR(r, t, p), j), ZX_MULS(v, 2));
    GEN y3 = ZX_SUB(
                FPXQ_MUL(r, ZX_SUB(v, x3), t, p),
                FPXQ_MUL(ZX_MULS(y1, 2), j, t, p)
             );
    GEN z3 = ZX_SUB(ZX_SUB(FPXQ_SQR(ZX_ADD(z1, h), t, p), z1z1), hh);
    
    //
    //    CL(P1P2) = z1.h.L(P1P2)
    //              = z1.h.(Y - y2) - (s2 - y1).(X - x2)
    //              = la - lb
    //
    GEN z1h = ZX_MUL(z1, h);

    GEN la = FPX_FP_ADD(FPX_NEG(y2, p), gel(P, 2), p);
        la = FPX_MUL(la, z1h, p);    
    GEN lb = FPX_FP_ADD(FPX_NEG(x2, p), gel(P, 1), p);
        lb = FPX_MUL(lb, ZX_SUB(s2, y1), p);
    
    GEN line = FPX_ADD(la, FPX_NEG(lb, p), p);
    
    pdata->num = FPXQ_MUL(pdata->num, line, t, p);

    gel(pdata->point, 1) = Rg_to_FpXQ(x3, t, p);
    gel(pdata->point, 2) = Rg_to_FpXQ(y3, t, p);
    gel(pdata->point, 3) = Rg_to_FpXQ(z3, t, p);
    
    if (!pdata->do_denom_elim) {
        pdata->denom = FPXQ_MUL(
                           pdata->denom,
                           eval_vert_jac_fpxq_fp(pdata, pdata->point, P),
                           t,
                           p
                       );
    }
    OUT();
}
//----------------------------------------------------------------------------
MAYBE_UNUSED GEN
std_miller_jac_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,P}(Q).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_jac_fp_fpxq(pdata, Q);

        if (_get_bit(absr, i)) {
            update_chord_jac_fp_fpxq(pdata, P, Q);
        }
        i--;
    }
    OUT();
    if (r_is_neg) {
        GEN vrp = eval_vert_jac_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        return pdata->num;
    }
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_jac_fp_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    //
    // Miller algorithm computing f_{r,P}(Q) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(P);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    GEN minusP = _get_opposite_point_jac_fp(P, p);
    
    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(r);

    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_jac_fp_fpxq(pdata, Q);

        if (naf[i] == 1) {
            update_chord_jac_fp_fpxq(pdata, P, Q);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{1, P} * v_{(i-1)P})
                //
                update_chord_jac_fp_fpxq(pdata, minusP, Q);
                
                if (!pdata->do_denom_elim) {
                    pdata->denom = FPXQ_MUL(
                                     pdata->denom,
                                     eval_vert_jac_fp_fpxq(pdata, P, Q),
                                     T,
                                     p
                                   );
                }
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        GEN vrp = eval_vert_jac_fp_fpxq(pdata, pdata->point, Q);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED  GEN
std_miller_jac_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Standard Miller algorithm computing f_{r,Q}(P).
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));

    GEN p = pdata->charac;
    GEN T = pdata->irredpol;

    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    long i = expi(absr) - 1;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_jac_fpxq_fp(pdata, P);
        if (_get_bit(absr, i)) {
            update_chord_jac_fpxq_fp(pdata, Q, P);
        }
        i--;
    }
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_jac_fpxq_fp(pdata, pdata->point, P);
        pdata->num = FPXQ_MUL(pdata->num, vrp, T, p);
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        OUT();
        if (pdata->do_denom_elim) {
            return FPXQ_INV(pdata->num, T, p);
        }
        return FPXQ_DIV(pdata->denom, pdata->num, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);
}
//----------------------------------------------------------------------------
GEN
naf_miller_jac_fpxq_fp(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    //
    // Miller algorithm computing f_{r,Q}(P) using a NAF representation of r.
    //
    // E is the elliptic curve.
    // r the order of the point P, as a t_INT.
    // Q a point on E(F_{p^k}) whose coordinates are given by ZXs.
    // P a point on E(F_p) whose coordinates are given by t_INTs.
    // T is the polynomial on F_p defining F_{p^k} given as a ZX.
    // p is the characteristic of the fields involved, as a t_INT.
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    pdata->point = gcopy(Q);
    
    setvarn(pdata->num, varn(gel(Q, 1)));
    setvarn(pdata->denom, varn(gel(Q, 1)));
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    
    GEN minusQ = _get_opposite_point_jac_fpxq(Q, p);

    int r_is_neg  = (cmpis(r, 0) < 0);
    GEN absr = absi(r);
    int* naf = _naf(absr);

    int  i = expi(absr) + 1;
    while ((naf[i] == 0) && (i > 0)) {
        i--;
    }
    i--;

    while (i >= 0) {
        pdata->num = FPXQ_SQR(pdata->num, T, p);
        if (!pdata->do_denom_elim) {
            pdata->denom = FPXQ_SQR(pdata->denom, T, p);
        }
        update_tangent_jac_fpxq_fp(pdata, P);

        if (naf[i] == 1) {
            update_chord_jac_fpxq_fp(pdata, Q, P);
        } else {
            if (naf[i] == -1) {
                //
                // f_{i-1, P} = f_{i, P} * l_{iP, -P} / (v_{1, P} * v_{(i-1)P})
                //
                update_chord_jac_fpxq_fp(pdata, minusQ, P);
                pdata->denom = FPXQ_MUL(
                                 pdata->denom,
                                 eval_vert_jac_fpxq_fp(pdata, Q, P),
                                 T,
                                 p
                               );
            }
        }
        i--;
    }
    free(naf);
    if (r_is_neg) {
        //
        // f_{-r, Q} = 1 / (f_{r, Q} * v_(r, Q))
        //
        GEN vrp = eval_vert_jac_fpxq_fp(pdata, pdata->point, P);
        GEN tmp = pdata->denom;
        pdata->denom = FPXQ_MUL(pdata->num, vrp, T, p);
        pdata->num   = tmp;
        gel(pdata->point, 2) = FPX_NEG(gel(pdata->point, 2), p);
        OUT();
        return FPXQ_DIV(pdata->num, pdata->denom, T, p);
    }
    if (pdata->do_denom_elim) {
        OUT();
        return pdata->num;
    }
    OUT();    
    return FPXQ_DIV(pdata->num, pdata->denom, T, p);

}
//----------------------------------------------------------------------------
